import logging
import time
import numpy as np
from kmer_mapper.mapper import read_fasta_into_chunks, map_kmers_to_graph_index
from kmer_mapper.util import remap_array, remap_array2
from scipy.ndimage import convolve1d
import pandas as pd
import scipy.signal
from graph_kmer_index import KmerIndex
from graph_kmer_index.shared_mem import from_shared_memory, to_shared_memory, SingleSharedArray
from pathos.multiprocessing import Pool
from itertools import repeat

def get_reads_as_matrices(read_file_name, chunk_size=500000, max_read_length=150):
    return (chunk for chunk in read_fasta_into_chunks(read_file_name, chunk_size, max_read_length))


def convert_byte_read_array_to_int_array(reads, read_length):
    # each item is a byesequence, one byte for each base pair
    return np.frombuffer(reads, dtype=np.uint8).reshape((len(reads), read_length))


def convert_read_matrix_to_numeric(read_matrix, give_complement_base_values=False):
    # from byte values
    from_values = np.array([65, 67, 71, 84, 97, 99, 103, 116], dtype=np.uint64)  # NB: Must be increasing
    # to internal base values for a, c, t, g
    to_values = np.array([0, 1, 3, 2, 0, 1, 3, 2], dtype=np.uint64)
    if give_complement_base_values:
        to_values = np.array([2, 3, 1, 0, 2, 3, 1, 0], dtype=np.uint64)

    return remap_array(read_matrix, from_values, to_values)


def get_kmer_hashes(numeric_read_matrix, is_complement=False, k=31):
    power_array = np.power(4, np.arange(0, k, dtype=np.uint64), dtype=np.uint64)
    if is_complement:
        power_array = np.power(4, np.arange(0, k, dtype=np.uint64)[::-1], dtype=np.uint64)

    k_half = k//2  # we don't want first k half and last k half columns
    assert numeric_read_matrix.dtype == np.uint64
    assert power_array.dtype == np.uint64

    #res = scipy.signal.fftconvolve(numeric_read_matrix, power_array, mode="valid")[:,k_half:-k_half]
    return convolve1d(numeric_read_matrix, power_array, mode="constant")[:,k_half:-k_half]

def get_kmer_hashes_numpy(numeric_read_matrix, is_complement=False, k=31):
    power_array = np.power(4, np.arange(0, k, dtype=np.uint64), dtype=np.uint64)
    if is_complement:
        power_array = np.power(4, np.arange(0, k, dtype=np.uint64)[::-1], dtype=np.uint64)

    assert numeric_read_matrix.dtype == np.uint64
    assert power_array.dtype == np.uint64

    # idea: convolve a 1d version of the 2d matrix. Use mode full and remove the first k-1 which are when power array is
    # convolved outside beginning. Reshape back to matrix and remove last k-1 columns which are overlaps between rows
    function = np.convolve
    #function = scipy.signal.convolve
    return function(numeric_read_matrix.flatten(), power_array, mode="full")[k-1:].reshape(numeric_read_matrix.shape)[:,:-(k-1)]


def get_unique_kmers_and_counts(hash_matrix):
    return np.unique(hash_matrix, return_counts=True)
    unique = pd.value_counts(hash_matrix, sort=False)
    return unique.keys(), unique.values


def get_kmers_from_read_matrix(read_matrix, mask, k=31, return_only_kmers=False, include_reverse_complement=True):
    logging.info("k=%d" % k)
    t = time.time()
    numeric_reads = convert_read_matrix_to_numeric(read_matrix)
    numeric_reads_complement = convert_read_matrix_to_numeric(read_matrix, True)
    logging.info("Time to convert reads to numeric: %.3f" % (time.time()-t))
    t = time.time()

    mask = mask[:,k-1:]  # make mask for hashes (we don't want hashes for bases not from reads, which are False in original mask)
    hashes = get_kmer_hashes_numpy(numeric_reads, k=k)[mask]
    if include_reverse_complement:
        hashes_complement = get_kmer_hashes_numpy(numeric_reads_complement, is_complement=True, k=k)[mask]
    logging.info("Time to get kmer hashes: %.3f" % (time.time()-t))
    t = time.time()

    if include_reverse_complement:
        all_hashes = np.concatenate([hashes, hashes_complement])
    else:
        all_hashes = hashes  # np.concatenate([hashes, hashes_complement])

    logging.info("N hashes in total: %d" % len(all_hashes))
    logging.info("Time to concatenate hashes and complement hashes: %.3f" % (time.time()-t))
    t = time.time()

    if return_only_kmers:
        return all_hashes

    unique, counts = get_unique_kmers_and_counts(all_hashes)
    #logging.info("N unique hashes  : %d" % len(unique))
    logging.info("Time to count kmers: %.3f" % (time.time()-t))

    return unique, counts


def get_kmers_from_fasta(fasta_file_name, chunk_size=500000, k=31, max_read_length=150, return_only_kmers=False):
    t = time.time()
    #logging.info("Read %d reads" % reads.shape[0])
    #logging.info("Time reading from file: %.4f" % (time.time()-t))
    reads, mask = next(get_reads_as_matrices(fasta_file_name, chunk_size=chunk_size, max_read_length=max_read_length))
    return get_kmers_from_read_matrix(reads, mask, k=k, return_only_kmers=return_only_kmers)


def map_fasta_single_thread(data):
    start_time = time.time()
    reads, args = data
    read_matrix, mask = reads
    if isinstance(read_matrix, str):
        read_matrix = from_shared_memory(SingleSharedArray, read_matrix).array
        mask = from_shared_memory(SingleSharedArray, mask).array

    t = time.time()
    read_matrix = convert_byte_read_array_to_int_array(read_matrix, args.max_read_length).astype(np.uint64)
    logging.info("Spent %.5f sec to convert byte read array to int read matrix " % (time.time()-t))

    shared_counts = from_shared_memory(SingleSharedArray, "counts_shared"+args.random_id).array

    index = from_shared_memory(KmerIndex, "kmer_index"+args.random_id)
    kmers = get_kmers_from_read_matrix(read_matrix, mask, args.kmer_size, True, not args.ignore_reverse_complement)
    node_counts = map_kmers_to_graph_index(index, args.n_nodes, kmers, args.max_hits_per_kmer)
    shared_counts += node_counts
    #shared_memory_name = "node_counts"+str(np.random.randint(0,10e15))
    #to_shared_memory(SingleSharedArray(node_counts), shared_memory_name)
    end_time = time.time()
    logging.info("One chunk took %.4f sec " % (end_time-start_time))
    #return shared_memory_name


def map_fasta(args):
    index = KmerIndex.from_file(args.kmer_index)
    args.random_id = str(np.random.random())
    to_shared_memory(index, "kmer_index"+args.random_id)
    n_nodes = index.max_node_id()
    args.n_nodes = n_nodes
    start_time = time.time()

    pool = Pool(args.n_threads)
    node_counts = np.zeros(n_nodes+1, dtype=float)

    to_shared_memory(SingleSharedArray(node_counts), "counts_shared" + args.random_id)

    reads = read_fasta_into_chunks(args.fasta_file, args.chunk_size, args.max_read_length, write_to_shared_memory=True, process_reads=False)
    logging.info("Got reads, starting processes")
    for result in pool.imap(map_fasta_single_thread, zip(reads, repeat(args))):
        continue

    node_counts = from_shared_memory(SingleSharedArray, "counts_shared"+args.random_id).array
    np.save(args.output_file, node_counts)
    logging.info("Saved node counts to %s.npy" % args.output_file)
    logging.info("Spent %.3f sec in total mapping kmers using %d threads" % (time.time()-start_time, args.n_threads))



#cython: language_level=3
from libc.stdio cimport *
import numpy as np
cimport numpy as np
import logging
import time
from kmer_mapper import util
cimport cython
import SharedArray as sa
from graph_kmer_index.shared_mem import to_shared_memory, SingleSharedArray

def py_read_file(filename):
    with open(filename, "r") as f:
        return f.read()

cpdef test():
    print("test2")


cpdef count_lines_in_fasta(file_name):
    i = 0
    with open(file_name) as f:
        #data = f.read()
        for line in f:
            i += 1

    return i

#@cython.boundscheck(False)
#@cython.wraparound(False)
cdef void read_sequence_to_kmer_hashes(char* sequence, np.uint64_t[:] read_array, np.uint64_t[:] reverse_complement_read_array,
                                       int sequence_length):
    #cdef sequence_length = len(sequence) - 3  # last two bytes are \n
    cdef int i = 0
    cdef int i_reverse_complement = sequence_length
    cdef int base
    for i in range(0, sequence_length):
        base = sequence[i]
        if base == 97 or base == 65:
            read_array[i] = 0
            reverse_complement_read_array[i_reverse_complement] = 2
        elif base == 99 or base == 67:
            read_array[i] = 1
            reverse_complement_read_array[i_reverse_complement] = 3
        elif base == 116 or base == 84:
            read_array[i] = 2
            reverse_complement_read_array[i_reverse_complement] = 0
        elif base == 103 or base == 71:
            read_array[i] = 3
            reverse_complement_read_array[i_reverse_complement] = 1
        else:
            # invalid characters, like N, just keep these 0
            read_array[i] = 0
        i_reverse_complement -= 1

def read_fasta(filename):
    filename_byte_string = filename.encode("UTF-8")
    cdef char * fname = filename_byte_string

    cdef FILE * cfile
    cfile = fopen(fname, "rb")
    if cfile == NULL:
        raise FileNotFoundError(2, "No such file or directory: '%s'" % filename)

    cdef char * line = NULL
    cdef size_t l = 0
    cdef ssize_t read
    cdef i = 0
    while True:
        i += 1
        read = getline(&line, &l, cfile)
        if read == -1: break

        # skip header lines
        if line[0] == 62:
            continue

        yield line

    fclose(cfile)


cdef void process_line(char* sequence, np.uint64_t[:] chunk_row, np.uint8_t[:] mask_row):
    cdef i = 0
    for i in range(0, len(sequence)-1):
        chunk_row[i] = sequence[i]
        mask_row[i] = True


def read_fasta_into_chunks(filename, chunk_size=1000, int max_read_length=150, write_to_shared_memory=False, process_reads=False):
    filename_byte_string = filename.encode("UTF-8")
    cdef char * fname = filename_byte_string

    cdef FILE * cfile
    cfile = fopen(fname, "rb")
    if cfile == NULL:
        raise FileNotFoundError(2, "No such file or directory: '%s'" % filename)

    cdef char * line = NULL
    cdef size_t l = 0
    cdef ssize_t read
    cdef i = 0
    cdef int line_length = 0

    if process_reads:
        chunk = np.zeros((chunk_size, max_read_length), dtype=np.uint64)
    else:
        #chunk = np.zeros((chunk_size, max_read_length), dtype=np.uint64)
        chunk = np.empty(chunk_size, dtype="|S" + str(max_read_length))  # max_read_length bytes for each element
        #chunk = np.empty(chunk_size, dtype="<U" + str(max_read_length))

    mask = np.zeros((chunk_size, max_read_length),
                        dtype=np.bool)  # True where reads have bases (for handling short reads)

    prev_time = time.time()

    while True:
        read = getline(&line, &l, cfile)
        if read == -1: break

        # skip header lines
        if line[0] == 62:
            continue

        line_length = len(line)-1
        #chunk[i,line_length] = line
        if process_reads:
            chunk[i, 0:line_length] = np.frombuffer(line, dtype=np.uint8)[:-1]  # remove \n at end
            mask[i, 0:line_length] = True
            # a bit slower:
            #process_line(line, chunk[i,:], mask[i,:])
        else:
            chunk[i] = line
            mask[i, 0:line_length] = True


        i += 1
        if i >= chunk_size:
            if write_to_shared_memory:
                name1 = "shared_array_" + str(np.random.random())
                name2 = "shared_array_" + str(np.random.random())
                to_shared_memory(SingleSharedArray(chunk), name1)
                to_shared_memory(SingleSharedArray(mask), name2)
                yield name1, name2
            else:
                yield chunk, mask
            logging.info("Took %.3f sec to read %d reads into matrices" % (time.time()-prev_time, chunk_size))
            prev_time = time.time()

            i = 0
            if process_reads:
                chunk = np.zeros((chunk_size, max_read_length), dtype=np.uint64)
            else:
                chunk = np.empty(chunk_size, dtype="|S" + str(max_read_length))

            mask = np.zeros((chunk_size, max_read_length), dtype=np.bool)  # True where reads have bases (for handling short reads)

    if write_to_shared_memory:
        name1 = "shared_array_" + str(np.random.random())
        name2 = "shared_array_" + str(np.random.random())
        to_shared_memory(SingleSharedArray(chunk[0:i]), name1)
        to_shared_memory(SingleSharedArray(mask[0:i]), name2)
        yield name1, name2
    else:
        yield chunk[0:i], mask[0:i]

    fclose(cfile)

cpdef handle_reads(filename):

    cdef int sum_hashes = 0
    cdef int max_read_length = 150
    #cdef np.uint64_t[:] read_array = np.zeros(read_length, dtype=np.uint64)
    cdef np.ndarray[np.uint64_t] read_array = np.zeros(max_read_length, dtype=np.uint64)
    #cdef np.uint64_t[:] reverse_complement_read_array = np.zeros(read_length, dtype=np.uint64)
    cdef np.ndarray[np.uint64_t] reverse_complement_read_array = np.zeros(max_read_length, dtype=np.uint64)
    cdef int k = 31
    cdef np.ndarray[np.int64_t] power_array = np.power(4, np.arange(0, k))
    cdef int read_length
    cdef int i = 0

    #lines = util.read_fasta(filename)
    #lines = util.read_fasta(filename)
    lines = read_fasta(filename)
    for line in lines:
        read_length = len(line) - 3
        read_sequence_to_kmer_hashes(line, read_array, reverse_complement_read_array, read_length)
        sum_hashes += read_array[0]
        kmers = np.convolve(read_array, power_array, mode='valid')
        reverse_complement_kmers = np.convolve(reverse_complement_read_array, power_array, mode='valid')
        sum_hashes += reverse_complement_read_array[10]

        if i == 0:
            print(kmers)

        i += 1

    print(sum_hashes)


cdef read_file(filename):
    cdef i = 0
    result = read_fasta(filename)
    return i


cpdef get_count(filename):
    return handle_reads(filename)



@cython.boundscheck(False)
@cython.wraparound(False)
def map_kmers_to_graph_index(index, int max_node_id, np.uint64_t[:] kmers, int max_index_lookup_frequency=1000):

    # index arrays
    cdef np.int64_t[:] hashes_to_index = index._hashes_to_index
    cdef np.uint32_t[:] n_kmers = index._n_kmers
    cdef np.uint32_t[:] nodes = index._nodes
    cdef np.uint64_t[:] ref_offsets = index._ref_offsets
    cdef np.uint64_t[:] index_kmers = index._kmers
    cdef np.uint16_t[:] index_frequencies = index._frequencies

    cdef int n_local_hits
    cdef unsigned long index_position

    cdef int l, j, i
    cdef long hash
    cdef int modulo = index._modulo
    #logging.info("Hash modulo is %d. Max index lookup frequency is %d." % (modulo, max_index_lookup_frequency))

    cdef np.ndarray[np.float_t] node_counts = np.zeros(max_node_id+1, dtype=np.float)
    #cdef np.float_t[:] node_counts = np.zeros(max_node_id+1, dtype=np.float)

    t = time.time()
    cdef np.uint64_t[:] kmer_hashes = np.mod(kmers, modulo)
    #logging.info("Time spent taking modulo: %.4f" % (time.time()-t))
    t = time.time()
    cdef int n_collisions = 0
    cdef int n_kmers_mapped = 0
    cdef int n_skipped_high_frequency = 0
    cdef int n_no_index_hits = 0
    #logging.info("Will process %d kmers" % kmers.shape[0])
    for i in range(kmers.shape[0]):
        hash = kmer_hashes[i]

        if hash == 0:
            continue

        n_local_hits = n_kmers[hash]
        if n_local_hits == 0:
            n_no_index_hits += 1

        index_position = hashes_to_index[hash]

        for j in range(n_local_hits):
            l = index_position + j
            # Check that this entry actually matches the kmer, sometimes it will not due to collision
            if index_kmers[l] != kmers[i]:
                n_collisions += 1
                continue

            if index_frequencies[l] > max_index_lookup_frequency:
                n_skipped_high_frequency += 1
                continue
            node_counts[nodes[l]] += 1
            n_kmers_mapped += 1

    logging.info("Time spent looking up kmers in index: %.3f" % (time.time()-t))
    #logging.info("N kmers with no index hits: %d" % n_no_index_hits)
    #logging.info("N hash collisions: %d" % n_collisions)
    #logging.info("N skipped because too high frequency: %d" % n_skipped_high_frequency)
    #logging.info("N kmers mapped: %d" % n_kmers_mapped)
    return node_counts


    


import logging
import sys
logging.basicConfig(stream=sys.stdout, level=logging.INFO, format='%(asctime)s %(levelname)s: %(message)s')
import numpy as np
import argparse
from graph_kmer_index import KmerIndex
from kmer_mapper.mapping import get_kmers_from_fasta, map_fasta
from kmer_mapper.mapper import map_kmers_to_graph_index
import time



def main():
    run_argument_parser(sys.argv[1:])

def test(args):
    print("Hi")


def map_fasta_command(args):
    map_fasta(args.kmer_index, args.fasta_file, args.chunk_size, args.n_threads, args.max_read_length, args.k)


def run_argument_parser(args):

    parser = argparse.ArgumentParser(
        description='Kmer Mapper',
        prog='kmer_mapper',
        formatter_class=lambda prog: argparse.HelpFormatter(prog, max_help_position=50, width=100))

    subparsers = parser.add_subparsers()
    subparser = subparsers.add_parser("map")
    subparser.add_argument("-i", "--kmer-index", required=True)
    subparser.add_argument("-f", "--fasta-file", required=True)
    subparser.add_argument("-k", "--kmer-size", required=False, default=31, type=int)
    subparser.add_argument("-t", "--n-threads", required=False, default=16, type=int)
    subparser.add_argument("-c", "--chunk-size", required=False, type=int, default=500000, help="N reads to process in each chunk")
    subparser.add_argument("-l", "--max-read-length", required=False, type=int, default=150,
                           help="Maximum length of reads. Reads should not be longer than this.")
    subparser.add_argument("-o", "--output-file", required=True)
    subparser.add_argument("-r", "--ignore-reverse-complement", required=False, default=False, type=bool)
    subparser.add_argument("-I", "--max-hits-per-kmer", required=False, default=1000, type=int,
                           help="Ignore kmers that have more than this amount of hits in index")
    subparser.set_defaults(func=map_fasta)


    if len(args) == 0:
        parser.print_help()
        sys.exit(1)

    args = parser.parse_args(args)
    args.func(args)
